/*
 * Copyright (C) 1999 Marek Michalkiewicz <marekm@linux.org.pl>
 * Copyright (C) 2000 Serge Vakulenko <vak@cronyx.ru>
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * without any conditions or restrictions.  This software is provided
 * "as is" without express or implied warranty.
 */
#ifndef _IO_H_
#define _IO_H_

#ifdef AVR_AT90S2313
#  include <machine/io2313.h>
#else
#ifdef AVR_AT90S2323
#  include <machine/io2323.h>
#else
#ifdef AVR_AT90S2333
#  include <machine/io2333.h>
#else
#ifdef AVR_AT90S2343
#  include <machine/io2343.h>
#else
#ifdef AVR_ATtiny22
#  include <machine/iotn22.h>
#else
#ifdef AVR_AT90S4414
#  include <machine/io4414.h>
#else
#ifdef AVR_AT90S4433
#  include <machine/io4433.h>
#else
#ifdef AVR_AT90S4434
#  include <machine/io4434.h>
#else
#ifdef AVR_AT90S8515
#  include <machine/io8515.h>
#else
#ifdef AVR_AT90C8534
#  include <machine/io8534.h>
#else
#ifdef AVR_AT90S8535
#  include <machine/io8535.h>
#else
#ifdef AVR_ATmega161
#  include <machine/iom161.h>
#else
#ifdef AVR_ATmega603
#  include <machine/iom603.h>
#else
#ifdef AVR_ATmega103
#  include <machine/iom103.h>
#endif /*  AVR_ATmega103 */
#endif /*  AVR_ATmega603 */
#endif /*  AVR_ATmega161 */
#endif /*  AVR_AT90S8535 */
#endif /*  AVR_AT90C8534 */
#endif /*  AVR_AT90S8515 */
#endif /*  AVR_AT90S4434 */
#endif /*  AVR_AT90S4433 */
#endif /*  AVR_AT90S4414 */
#endif /*  AVR_ATtiny22 */
#endif /*  AVR_AT90S2343 */
#endif /*  AVR_AT90S2333 */
#endif /*  AVR_AT90S2323 */
#endif /*  AVR_AT90S2313 */

#ifndef __ASSEMBLER__

/*
 * Global interrupt control.
 */
#define sei()  asm volatile ("sei" ::: "memory")
#define cli()  asm volatile ("cli" ::: "memory")

/*
 * Compile a string constant into the program (flash) memory.
 */
#define __flash__ __attribute__ ((progmem))
#define CONST(s) ({static char c[] __flash__ = s; c;})

/*
 * Read a byte from the program space (flash).
 * Only low 64 kbytes are addressable.
 */
#define readb(addr) ({					\
	register unsigned char t;			\
	asm volatile (					\
		"lpm" "\n"				\
		"mov	%0,r0"				\
		: "=r" (t)				\
		: "z" ((int)addr));			\
	t; })

/*
 * Memory mapped I/O - use for non-constant I/O space addresses.
 */
#define __mmiob(port) (*((volatile unsigned char*)((char)(port) + 0x20)))

#define __mminw(port) (*((volatile unsigned short*)((char)(port) + 0x20)))

#define __mmoutw(val, port)				\
	asm volatile (					\
		"std Z+1, %B0" "\n" /* MSB first */	\
		"st Z, %A0"				\
		: /* no outputs */			\
		: "r" ((short)(val)),			\
		  "z" ((char)(port) + 0x20))

/*
 * Dumb versions of I/O macros.
 * The order of parameters is taken from Linux.
 */
#define inb(port)		__mmiob (port)
#define outb(val, port)		do { __mmiob (port) = (val); } while (0)
#define inw(port)		__mminw (port)
#define outw(val, port)		__mmoutw ((val), (port))
#define setb(bit, port)		do { __mmiob (port) |= 1 << (bit); } while (0)
#define clearb(bit, port)	do { __mmiob (port) &= ~(1 << (bit)); } while (0)
#define testb(bit, port)	(__mmiob (port) & (1 << (bit)))

/*
 * Macros for access to 8-bit I/O registers with constant address.
 * The order of parameters is as in Linux.
 */
#define outb_const(val, port)				\
	asm volatile (					\
		"out %1, %0"				\
		: /* no outputs */			\
		: "r" ((char)(val)),			\
		  "I" ((char)(port)))

#define outb_const_zero(port)				\
	asm volatile (					\
		"out %0, __zero_reg__"			\
		: /* no outputs */			\
		: "I" ((char)(port)))

#define inb_const(port) ({				\
	register unsigned char t;			\
	asm volatile (					\
		"in %0, %1"				\
		: "=r" (t)				\
		: "I" ((char)(port)));			\
	t; })

/*
 * Macros for access to 16-bit I/O registers (ADC, ICR1, OCR1, TCNT1) -
 * read low byte first, then high byte (from the TEMP register),
 * write high byte first (to the TEMP register), then low byte.
 */
#define outw_const(val, port)				\
	asm volatile (					\
		"out (%1)+1, %B0" "\n"			\
		"out %1, %A0"				\
		: /* no outputs */			\
		: "r" ((short)(val)),			\
		  "I" ((char)(port)))

#define inw_const(port) ({				\
	register unsigned short t;			\
	asm volatile (					\
		"in %A0, %1" "\n"			\
		"in %B0, (%1)+1"			\
		: "=r" (t)				\
		: "I" ((char)(port)));			\
	t; })

/*
 * Bit handling macros.
 */
#define setb_const(bit, port) 				\
	asm volatile (					\
		"sbi %0, %1"				\
		: /* no outputs */			\
		: "I" ((char)(port)),			\
		  "I" ((char)(bit)))

#define clearb_const(bit, port) 			\
	asm volatile (					\
		"cbi %0, %1"				\
		: /* no outputs */			\
		: "I" ((char)(port)),			\
		  "I" ((char)(bit)))

#define testb_const(bit, port) ({			\
	register unsigned char t;			\
	asm volatile (					\
		"clr %0" "\n"				\
		"sbic %1, %2" "\n"			\
		"inc %0"				\
		: "=r" (t)				\
		: "I" ((char)(port)),			\
		  "I" ((char)(bit)));			\
	t; })

/*
 * Optimized versions of I/O macros.
 */
#ifdef __OPTIMIZE__
#undef inb
#undef outb
#undef inw
#undef outw
#undef setb
#undef clearb
#undef testb

#define inb(port) (__builtin_constant_p ((port)) ?	\
	inb_const ((port)) : __mmiob ((port)))

#define outb(val, port) do {				\
	if (! __builtin_constant_p ((port)))		\
		__mmiob ((port)) = (val);		\
	else if (__builtin_constant_p ((val)) && (val) == 0) \
		outb_const_zero ((port));		\
	else						\
		outb_const ((val), (port));		\
	} while (0)

#define inw(port) (__builtin_constant_p (port) ?	\
	inw_const (port) : __mminw (port))

#define outw(val, port) do {				\
	if (! __builtin_constant_p ((port)))		\
		__mmoutw ((val), (port));		\
	else						\
		outw_const ((val), (port));		\
	} while (0)

#define setb(bit, port) do {				\
	if (! __builtin_constant_p ((port)))		\
		__mmiob ((port)) |= 1 << (bit);		\
	else if (! __builtin_constant_p ((bit)))	\
		asm volatile (				\
		"in __tmp_reg__, %0" "\n"		\
		"or __tmp_reg__, %1" "\n"		\
		"out %0, __tmp_reg__"			\
		: /* no outputs */			\
		: "I" ((char)(port)),			\
		  "r" ((char)1 << (bit)));		\
	else if ((port) >= 0x20) {			\
		register char dummy;			\
		asm volatile (				\
		"in %0, %1" "\n"			\
		"ori %0, %2" "\n"			\
		"out %1, %0"				\
		: "=r" (dummy)				\
		: "I" ((char)(port)),			\
		  "M" ((unsigned char)1 << (bit)));	\
	} else						\
		setb_const (bit, port);			\
	} while (0)

#define clearb(bit, port) do {				\
	if (! __builtin_constant_p(port))		\
		__mmiob ((port)) &= ~(1 << (bit));	\
	else if (! __builtin_constant_p(bit))		\
		asm volatile (				\
		"in __tmp_reg__, %0" "\n"		\
		"and __tmp_reg__, %1" "\n"		\
		"out %0, __tmp_reg__"			\
		: /* no outputs */			\
		: "I" ((char)(port)),			\
		  "r" ((char)~((char)1 << (bit))));	\
	else if ((port) >= 0x20) {			\
		register char dummy;			\
		asm volatile (				\
		"in %0, %1" "\n"			\
		"andi %0, %2" "\n"			\
		"out %1, %0"				\
		: "=r" (dummy)				\
		: "I" ((char)(port)),			\
		  "M" (0xff & ~(1 << (bit))));		\
	} else						\
		clearb_const (bit, port);		\
	} while (0)

#define testb(bit, port) ({				\
	unsigned char t;				\
	if (! __builtin_constant_p((port)))		\
		t = (__mmiob ((port)) & (1 << (bit))) != 0; \
	else if (! __builtin_constant_p((bit)))		\
		t = (inb_const (port) & (1 << (bit))) != 0; \
	else if ((port) >= 0x20)			\
		asm volatile (				\
		"clr %0" "\n"				\
		"in __tmp_reg__, %1" "\n"		\
		"sbrc __tmp_reg__, %2" "\n"		\
		"inc %0"				\
		: "=r" (t)				\
		: "I" ((char)(port)),			\
		  "I" ((char)(bit)));			\
	else						\
		t = testb_const (bit, port);		\
	t; })

#endif /*__OPTIMIZE__*/
#endif /* __ASSEMBLER__ */
#endif /* _IO_H_ */
