#include <sys/lib.h>

static int debug_char;

/*
 * Send a byte to the UART transmitter, with interrupts disabled.
 */
void
debug_putchar (void *arg, char c)
{
	char x;

	x = inb (SREG);
	cli();

	if (! inb (UBRR))
		outb (((int) (KHZ * 1000L / 9600) + 8) / 16 - 1, UBRR);
	setb (TXEN, UCR);

	/* Wait for transmitter idle. */
	while (! testb (UDRE, USR))
		continue;
again:
	/* Send byte. */
	outb (c, UDR);

	/* Wait for transmitter idle. */
	while (! testb (UDRE, USR))
		continue;

	if (c == '\n') {
		c = '\r';
		goto again;
	}

#ifndef NDEBUG
	if (testb (RXC, USR)) {
		debug_char = inb (UDR);
		if (debug_char == 3) {
			debug_char = -1;
			breakpoint ();
		}
	}
#endif
	outb (x, SREG);
}

/*
 * Wait for the byte to be received and return it.
 */
unsigned char
debug_getchar (void)
{
	unsigned char c;
	char x;

	if (debug_char >= 0) {
		c = debug_char;
		debug_char = -1;
		return c;
	}

	x = inb (SREG);
	cli();

	if (! inb (UBRR))
		outb (((int) (KHZ * 1000L / 9600) + 8) / 16 - 1, UBRR);
	setb (RXEN, UCR);
	for (;;) {
		/* Wait until receive data available. */
		while (! testb (RXC, USR))
			continue;
		c = inb (UDR);
#ifndef NDEBUG
		if (c == 3) {
			breakpoint ();
			continue;
		}
#endif
		break;
	}
	outb (x, SREG);
	return c;
}

void
debug_puts (const char *p)
{
	char c;

	for (;;) {
		c = readb ((int) p);
		if (! c)
			return;
		debug_putchar (0, c);
		++p;
	}
}
