/*
 * Machine-dependent uOS declarations for: Atmel AVR, GCC.
 */
#include <machine/io.h>

/*
 * Build the initial task's stack frame.
 * Arguments:
 *	t  - the task object, with the stack space appended
 *	f  - the task function to call
 *	a  - the function argument
 *	sz - stack size in bytes
 */
#define MACHDEP_BUILD_STACK_FRAME(t,f,a,sz) avr_build_stack_frame (t, \
	(unsigned short) f, (unsigned short) a, (char*)t + sz-1)
void avr_build_stack_frame (task_t *t, unsigned short func,
	unsigned short arg, char *stack);

/*
 * Perform the task switch.
 */
#define MACHDEP_TASK_SWITCH()	avr_task_switch()
void avr_task_switch (void);

/*
 * The total number of different hardware interrupts.
 */
#define MACHDEP_INTERRUPTS	26

/*
 * The global interrupt control.
 * Disable and restore the hardware interrupts,
 * saving the interrupt enable flag into the supplied variable.
 */
#define MACHDEP_INTR_DISABLE(x)	{ *(char*)(x) = inb (SREG); cli (); }
#define MACHDEP_INTR_RESTORE(x)	{ outb ((char) (x), SREG); }

/*
 * Allow the given hardware interrupt,
 * unmasking it in the interrupt controller.
 */
#define MACHDEP_INTR_ALLOW(n)	avr_intr_allow (n)
void avr_intr_allow (unsigned char irq);

/*
 * Idle system activity.
 * Enable interrupts and enter sleep mode.
 * (optional feature)
 */
#define MACHDEP_IDLE()		{ setb (SE, MCUCR);\
				asm volatile ("sei\n sleep\n nop"); }

/*
 * Uncomment these lines for remote debugging.
 */
#ifndef NDEBUG
#define MACHDEP_HALT()		breakpoint()
void breakpoint (void);
#endif
