/*
 * Machine-dependent part of uOS for: Intel i86, DOS, Borland Turbo C++.
 */
#include "sys/uos.h"
#include "sys/internal.h"
#include "machdep.h"

#define VECTOR_IRQ0_CHAIN 	0x62
#define VECTOR(irq)		(irq<8 ? irq+8 : irq+0x68)

static void interrupt (*saved_vector[16]) ();

/* Switch to the most priority task. */
#define SWITCH_TASK() {\
	/* Save current task stack. */\
	((short*)&task_current->stack_pointer)[0] = _SP;\
	((short*)&task_current->stack_pointer)[1] = _SS;\
	/* Compute new active task. */\
	task_policy ();\
	/* Switch to the new task. */\
	_SP = ((short*)&task_current->stack_pointer)[0];\
	_SS = ((short*)&task_current->stack_pointer)[1]; }

/*
 * Perform the task switch.
 */
void
interrupt dos_task_switch (void)
{
	disable();
	SWITCH_TASK();
}

/*
 * The interrupt handler pattern.
 */
#define HANDLE(n,action)\
static void interrupt intr_##n (void)\
{\
	action;\
	lock_activate (uos_handler[n], 0);\
	SWITCH_TASK();\
}

HANDLE (0, geninterrupt (VECTOR_IRQ0_CHAIN))
HANDLE (1, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x02))
HANDLE (2, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x04))
HANDLE (3, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x08))
HANDLE (4, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x10))
HANDLE (5, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x20))
HANDLE (6, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x40))
HANDLE (7, outportb (0x20, 0x20); outportb (0x21, inportb (0x21) | 0x80))
HANDLE (8, outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x01))
HANDLE (9, outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x02))
HANDLE (10,outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x04))
HANDLE (11,outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x08))
HANDLE (12,outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x10))
HANDLE (13,outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x20))
HANDLE (14,outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x40))
HANDLE (15,outportb (0xa0, 0x20); outportb (0x20, 0x20);\
	   outportb (0xa1, inportb (0xa1) | 0x80))

/*
 * Bind the handler to the given hardware interrupt.
 */
void
dos_intr_bind (unsigned char irq)
{
	static void *vector [16] = {
		intr_0,	 intr_1,  intr_2,  intr_3,
		intr_4,  intr_5,  intr_6,  intr_7,
		intr_8,  intr_9,  intr_10, intr_11,
		intr_12, intr_13, intr_14, intr_15,
	};

	saved_vector [irq] = getvect (VECTOR (irq));
	/* cprintf ("bind irq %d, save vector 0x%lx\r\n", irq, saved_vector [irq]); */
	setvect (VECTOR (irq), (void interrupt (*)()) vector [irq]);
	if (irq == 0)
		setvect (VECTOR_IRQ0_CHAIN, saved_vector [0]);
}

/*
 * Unbind the interrupt handler.
 */
void
dos_intr_unbind (unsigned char irq)
{
	/* cprintf ("unbind irq %d to 0x%lx\r\n", irq, saved_vector [irq]); */
	if (saved_vector [irq])
		setvect (VECTOR (irq), saved_vector [irq]);
}

/*
 * Allow the given hardware interrupt,
 * unmasking it in the interrupt controller.
 */
void
dos_intr_allow (unsigned char irq)
{
	/* cprintf ("allow irq %d\r\n", irq); */
	if (irq < 8)
		outportb (0x21, inportb (0x21) & ~(1 << irq));
	else
		outportb (0xa1, inportb (0xa1) & ~(1 << (irq & 7)));
}

/*
 * Halt the system: unbind all interrupts and exit to DOS.
 */
void
dos_halt (void)
{
	int irq;

	/* Restore all interrupt handlers. */
	for (irq=0; irq<16; ++irq)
		dos_intr_unbind (irq);
	enable ();
	exit (0);
}

/*
 * Build the initial task's stack frame.
 * Arguments:
 * t	- the task object
 * func	- the task function to call
 * arg	- the function argument
 * sp	- pointer to the (end of) stack space
 */
void
dos_build_stack_frame (task_t *t, void *func, void *arg, unsigned short *sp)
{
	*--sp = FP_SEG (arg);		/* task argument segment... */
	*--sp = FP_OFF (arg);           /* ...and pointer */
	*--sp = FP_SEG (func);		/* CS */
	*--sp = FP_OFF (func);		/* IP */
	*--sp = _FLAGS | 0x200;		/* flags - enable interrupts */
	*--sp = FP_SEG (func);		/* CS */
	*--sp = FP_OFF (func);		/* IP */
	*--sp = 0;			/* ax */
	*--sp = 0;			/* bx */
	*--sp = 0;			/* cx */
	*--sp = 0;			/* dx */
	*--sp = _ES;			/* es */
	*--sp = _DS;			/* ds */
	*--sp = 0;			/* si */
	*--sp = 0;			/* di */
	*--sp = 0;			/* bp */
	t->stack_pointer = sp;
}
