/*
 * Machine-dependent uOS declarations for: Intel i86, DOS, Borland Turbo C++.
 */
#include <dos.h>

/*
 * Build the initial task's stack frame.
 * Arguments:
 *	t  - the task object, with the stack space appended
 *	f  - the task function to call
 *	a  - the function argument
 *	sz - stack size in bytes
 */
#define MACHDEP_BUILD_STACK_FRAME(t,f,a,sz) \
	dos_build_stack_frame (t, (void*) f, a, (unsigned short*)t + (sz)/2)
void dos_build_stack_frame (task_t *t, void *func, void *arg,
	unsigned short *sp);

/*
 * Perform the task switch.
 */
#define MACHDEP_TASK_SWITCH()	dos_task_switch()
void interrupt dos_task_switch (void);

/*
 * The total number of different hardware interrupts.
 */
#define MACHDEP_INTERRUPTS	16

/*
 * The global interrupt control.
 * Disable and restore the hardware interrupts,
 * saving the interrupt enable flag into the supplied variable.
 */
#define MACHDEP_INTR_DISABLE(x)	{ asm { pushf; pop AX; cli }; *(x) = _AX; }
#define MACHDEP_INTR_RESTORE(x)	{ _AX = (x); asm { push AX; popf }; }

/*
 * Allow the given hardware interrupt,
 * unmasking it in the interrupt controller.
 */
#define MACHDEP_INTR_ALLOW(n)	dos_intr_allow (n)
void dos_intr_allow (unsigned char irq);

/*
 * Bind the handler to the given hardware interrupt.
 * (optional feature)
 */
#define MACHDEP_INTR_BIND(n)	dos_intr_bind (n)
void dos_intr_bind (unsigned char irq);

/*
 * Unbind the interrupt handler.
 * (optional feature)
 */
#define MACHDEP_INTR_UNBIND(n)	dos_intr_unbind (n)
void dos_intr_unbind (unsigned char irq);

/*
 * Idle system activity.
 * Enable interrupts and enter sleep mode.
 * (optional feature)
 */
#define MACHDEP_IDLE()		asm { sti; hlt }

/*
 * Halt the system: unbind all interrupts and exit to DOS.
 * (optional feature)
 */
#define MACHDEP_HALT()		dos_halt()
