#include "sys/uos.h"
#include "sys/nvram.h"

#if __MSDOS__
#endif

#if __AVR__
#   include <machine/io.h>
#   define NVRAM_IRQ		21	/* EEPROM write complete */
#endif

/*
 * Write a byte to NVRAM.
 */
void
nvram_write (nvram_t *v, int addr, char c)
{
	lock_acquire (&v->lock);

	/* Wait until EEWE becomes zero. */
	while (testb (EEWE, EECR))
		lock_wait (&v->lock);

	outw (addr, EEAR);
	outb (c, EEDR);
	setb_const (EEMWE, EECR);
	setb_const (EEWE, EECR);

	/* Wait until EEWE becomes zero. */
	while (testb (EEWE, EECR))
		lock_wait (&v->lock);

	lock_release (&v->lock);
}

/*
 * Read a byte from NVRAM.
 */
unsigned char
nvram_read (nvram_t *v, int addr)
{
	unsigned char c;

	lock_acquire (&v->lock);

	/* Wait until EEWE becomes zero. */
	while (testb (EEWE, EECR))
		lock_wait (&v->lock);

	outw (addr, EEAR);
	setb (EERE, EECR);

	/* Wait until EERE becomes zero.
	 * It could not take us too long. */
	while (testb (EERE, EECR))
		continue;

	c = inb (EEDR);
	lock_release (&v->lock);
	return c;
}

void
nvram_init (nvram_t *v)
{
	/* Get the interrupt. */
	lock_acquire_irq (&v->lock, NVRAM_IRQ);
	lock_release (&v->lock);
}
