#include "sys/uos.h"
#include "sys/timer.h"

#if __MSDOS__
#   define MSEC_PER_TICK	55	/* Standard MS-DOS timer */
#   define TIMER_IRQ		0	/* IRQ0 */
#endif

#if __AVR__
#   include <machine/io.h>
#   if KHZ > 6000
#      define MSEC_PER_TICK	2	/* Fast processor */
#   else
#      define MSEC_PER_TICK	10	/* Slow processor */
#   endif
#   define TIMER_IRQ		11	/* Timer 1 compare A */
#endif

/*
 * System timer task.
 */
static void
main_timer (void *arg)
{
	timer_t *t = arg;

	/* Request the IRQ. */
	lock_acquire_irq (&t->lock, TIMER_IRQ);

	/* Initialize the hardware. */
#if __AVR__
	outb (0, TCCR1A);
	outb (0, TCCR1B);
	outw ((unsigned) KHZ * MSEC_PER_TICK - 2, OCR1A);
	outw (0, TCNT1);
	outb (0x09, TCCR1B);	/* clock source CK, clear on match A */
#endif
	for (;;) {
		lock_wait (&t->lock);
		t->milliseconds += MSEC_PER_TICK;
	}
}

/*
 * Return the (real) time in milliseconds since uOS start.
 */
unsigned long
timer_milliseconds (timer_t *t)
{
	unsigned long val;

	lock_acquire (&t->lock);
	val = t->milliseconds;
	lock_release (&t->lock);
	return val;
}

/*
 * Delay the current task by the given time in milliseconds.
 */
void
timer_delay (timer_t *t, unsigned long msec)
{
	lock_acquire (&t->lock);
	msec += t->milliseconds;
	while (t->milliseconds < msec)
		lock_wait (&t->lock);
	lock_release (&t->lock);
}

void
timer_init (timer_t *t, int prio)
{
	/* Create timer task. */
	task_create (main_timer, t, prio, t->stack, sizeof (t->stack), 0);
}
