#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/machdep.h"
#include "sys/internal.h"

/*
 * Get the lock. The calling task would block
 * waiting for the lock to be released by somebody.
 * Associate IRQ with the lock, disable interrupts.
 * On interrupt, the signal message is sent to the lock.
 */
void
lock_acquire_irq (lock_t *m, int irq)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	while (m->master && m->master != task_current) {
		/* Monitor is locked, block the task. */
		assert (task_current->lock == 0);
		task_current->lock = m;
		task_dequeue (&task_active, task_current);
		task_enqueue (&m->slaves, task_current);
		MACHDEP_TASK_SWITCH ();
	}
	if (! m->master) {
		assert (m->slaves.head == 0);
		m->master = task_current;
		lock_enqueue (&task_current->slaves, m);
	}
#ifdef MACHDEP_INTR_BIND
	if (! uos_handler [irq])
		MACHDEP_INTR_BIND (irq);
#endif
	uos_handler [irq] = m;
	m->irq = 1 + irq;
	m->context = x;
}
