#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/machdep.h"
#include "sys/internal.h"

/*
 * Release the lock. All tasks waiting for the lock
 * are unblocked, possibly causing task switch.
 * Disassociate IRQ with the lock, enable interrupts.
 */
void
lock_release_irq (lock_t *m)
{
	int irq;

	assert (STACK_GUARD (task_current));
	if (m->irq) {
		irq = m->irq - 1;
		uos_handler [irq] = 0;
#ifdef MACHDEP_INTR_UNBIND
		MACHDEP_INTR_UNBIND (irq);
#endif
	}
	assert (m->master != 0);
	lock_dequeue (&m->master->slaves, m);
	m->master = 0;
	if (! m->slaves.head) {
		MACHDEP_INTR_RESTORE (m->context);
		return;
	}
	do {
		task_t *t = m->slaves.head;
		assert (t->lock == m);
		t->lock = 0;
		task_dequeue (&m->slaves, t);
		task_enqueue (&task_active, t);
	} while (m->slaves.head);
	MACHDEP_INTR_RESTORE (m->context);
	MACHDEP_TASK_SWITCH ();
}
