#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/machdep.h"
#include "sys/internal.h"

/*
 * Send the signal to the lock. All tasks waiting for the signal
 * on the lock are unblocked, possibly causing task switch.
 */
void
lock_signal (lock_t *m, void *message)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	if (! m->waiters.head) {
		MACHDEP_INTR_RESTORE (x);
		return;
	}
	lock_activate (m, message);
	MACHDEP_INTR_RESTORE (x);
	MACHDEP_TASK_SWITCH ();
}

/*
 * Wait for the signal on the lock. The calling task is blocked.
 * In the case the lock has associated IRQ number,
 * it is unmasked (permitted to happen).
 */
void *
lock_wait (lock_t *m)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	assert (task_current->wait == 0);
	task_current->wait = m;
	task_dequeue (&task_active, task_current);
	task_enqueue (&m->waiters, task_current);

 	if (m->irq)
 		MACHDEP_INTR_ALLOW (m->irq - 1);

 	if (! m->master) {
 		MACHDEP_INTR_RESTORE (x);
		MACHDEP_TASK_SWITCH ();
 		return task_current->message;
 	}

 	/* The lock is hold by the current task - release it. */
 	assert (m->master == task_current);
 	assert (task_current->slaves.head == m);
	lock_dequeue (&m->master->slaves, m);
	m->master = 0;
	task_current->context = m->context;
	while (m->slaves.head) {
		task_t *t = m->slaves.head;
		assert (t->lock == m);
		t->lock = 0;
		task_dequeue (&m->slaves, t);
		task_enqueue (&task_active, t);
	}
	MACHDEP_TASK_SWITCH ();

	/* Acquire the lock again. */
	while (m->master) {
		/* Monitor is locked, block the task. */
		assert (task_current->lock == 0);
		task_current->lock = m;
		task_dequeue (&task_active, task_current);
		task_enqueue (&m->slaves, task_current);
		MACHDEP_TASK_SWITCH ();
	}
	m->master = task_current;
	m->context = task_current->context;
 	lock_enqueue (&task_current->slaves, m);

	MACHDEP_INTR_RESTORE (x);
	return task_current->message;
}
