#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/machdep.h"
#include "sys/internal.h"

/*
 * Try to get the lock. Return 1 on success, 0 on failure.
 * The calling task does not block.
 * In the case the lock has associated IRQ number,
 * after acquiring the lock the IRQ will be disabled.
 */
int
lock_try (lock_t *m)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
 	if (m->master && m->master != task_current) {
 		/* Monitor is locked. */
 		MACHDEP_INTR_RESTORE (x);
 		return 0;
 	}
 	if (! m->master) {
 		assert (m->slaves.head == 0);
 		m->master = task_current;
		lock_enqueue (&task_current->slaves, m);
	}
	if (m->irq)
		m->context = x;
	else
		MACHDEP_INTR_RESTORE (x);
 	return 1;
}
